/* $NetBSD: ep93xx_intr.c,v 1.3 2005/01/05 04:53:50 joff Exp $ */

/*
 * Copyright (c) 2002 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jesse Off
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Ichiro FUKUHARA and Naoto Shimazaki.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $NetBSD: ep93xx_intr.c,v 1.3 2005/01/05 04:53:50 joff Exp $
 *
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD");

/*
 * Interrupt support for the Cirrus Logic EP93XX
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#define _ARM32_BUS_DMA_PRIVATE
#define __RMAN_RESOURCE_VISIBLE
#include <sys/rman.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <machine/cpufunc.h>

#include <arm/ep93xx/ep93xxreg.h> 
#include <arm/ep93xx/ep93xxvar.h> 

/* Software copy of the IRQs we have enabled. */
__volatile u_int32_t vic1_intr_enabled;
__volatile u_int32_t vic2_intr_enabled;

#define VIC1REG(reg)	*((volatile u_int32_t*) (EP93XX_AHB_VBASE + \
	EP93XX_AHB_VIC1 + (reg)))
#define VIC2REG(reg)	*((volatile u_int32_t*) (EP93XX_AHB_VBASE + \
	EP93XX_AHB_VIC2 + (reg)))


void
arm_unmask_irqs(int irqmask)
{
	int irq;

	while ((irq = ffs(irqmask) - 1)) {
		if (irq < VIC_NIRQ) {
			vic1_intr_enabled |= (1U << irq);
			VIC1REG(EP93XX_VIC_IntEnable) = (1U << irq);
		} else {
			vic2_intr_enabled |= (1U << (irq - VIC_NIRQ));
			VIC2REG(EP93XX_VIC_IntEnable) = 
			    (1U << (irq - VIC_NIRQ));
		}
	}
}

void
arm_mask_irqs(int irqmask)
{
	int irq;

	while ((irq = ffs(irqmask) - 1)) {
		if (irq < VIC_NIRQ) {
			vic1_intr_enabled &= ~(1U << irq);
			VIC1REG(EP93XX_VIC_IntEnClear) = (1U << irq);
		} else {
			vic2_intr_enabled &= ~(1U << (irq - VIC_NIRQ));
			VIC2REG(EP93XX_VIC_IntEnClear) = 
			    (1U << (irq - VIC_NIRQ));
		}
	}
}

int
arm_get_irqnb(void *unused)
{
	
	return (VIC1REG(EP93XX_VIC_IRQStatus) |
	    VIC2REG(EP93XX_VIC_IRQStatus));
}

static void
ep93xx_intr_init(void)
{

	vic1_intr_enabled = 0;
	vic2_intr_enabled = 0;


	/* All interrupts should use IRQ not FIQ */
	VIC1REG(EP93XX_VIC_IntSelect) = 0;
	VIC2REG(EP93XX_VIC_IntSelect) = 0;

	/* Enable IRQs (don't yet use FIQs). */
	enable_interrupts(I32_bit);
}

int	ep93xx_probe(device_t);
void	ep93xx_identify(driver_t *, device_t);
int	ep93xx_attach(device_t);

int
ep93xx_probe(device_t dev)
{
	device_set_desc(dev, "Cirrus Logic EP93xx CPU");
	return (0);
}

void
ep93xx_identify(driver_t *driver, device_t parent)
{
	
	BUS_ADD_CHILD(parent, 0, "ep", 0);
}

struct arm32_dma_range *
bus_dma_get_range(void)
{

	return (NULL);
}

int
bus_dma_get_range_nb(void)
{

	return (0);
}

int
ep93xx_attach(device_t dev)
{

	ep93xx_intr_init();
	device_add_child(dev, "uart", 0);
	device_add_child(dev, "epclk", 0);
	return (0);
}

void
cpu_reset(void)
{
	while(1);
}

static struct resource *
ep93xx_alloc_resource(device_t dev, device_t child, int type, int *rid,
    u_long start, u_long end, u_long count, u_int flags)
{
	if (type == SYS_RES_IRQ) {
		struct resource *res = malloc(sizeof(*res), M_DEVBUF, M_WAITOK);
		res->r_start = start;
		res->r_end = end;
		return (res);
	}
	return (NULL);
}

static int
ep93xx_setup_intr(device_t dev, device_t child,
    struct resource *ires, int flags, driver_intr_t *intr, void *arg,
    void **cookiep)
{
	BUS_SETUP_INTR(device_get_parent(dev), child, ires, flags, intr, arg,
	    cookiep);
	arm_unmask_irqs(1 << ires->r_start);
	
	return (0);
}
static device_method_t ep93xx_methods[] = {
	DEVMETHOD(device_probe, ep93xx_probe),
	DEVMETHOD(device_attach, ep93xx_attach),
	DEVMETHOD(device_identify, ep93xx_identify),
	DEVMETHOD(bus_alloc_resource, ep93xx_alloc_resource),
	DEVMETHOD(bus_setup_intr, ep93xx_setup_intr),
	{0, 0},
};

static driver_t ep93xx_driver = {
	"eparm",
	ep93xx_methods,
	1,
};
static devclass_t ep93xx_devclass;

DRIVER_MODULE(eparm, nexus, ep93xx_driver, ep93xx_devclass, 0, 0);
